/*  Binary Tree Data Structure Example
    Copyright (C) 2002  Clinton Ebadi

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "bintree.h"
#include <assert.h>
#include <stdlib.h> /* malloc */

typedef struct _node node; /* users shouldn't use nodes, so typedef is here */

/* some static functions used for convienience */

static node* new_node (); /* returns a new node with its members set to NULL */
/* returns a free node in the tree where dat should go using cmp_func f 
   NOTE: does not actually put dat into the tree */
static node* free_loc (node* tree_root, cmp_func f, void* dat);
/* same as free_loc but returns the location of dat if found */
static node* find_dat (node* tree_root, cmp_func f, void* dat);
/* removes dat from the tree */
static node* del_dat (node* nd, node* parent, cmp_func f, void* dat);


/* main functions */
bintree*
make_bintree (copy_func cp, cmp_func cm)
{
  bintree* temp = malloc (sizeof (bintree));
  temp->copy = cp;
  temp->cmp = cm;
  temp->tree = NULL; /* first insertion creates the tree */

  return temp;
}

void
bintree_insert (bintree* btree, void* dat)
{
  node* tnode;
  assert (btree != NULL);

  if (btree->tree == NULL) /* empty tree */
    {
      btree->tree = malloc (sizeof (node));
      btree->tree->left = NULL;
      btree->tree->right = NULL;
      btree->tree->data = btree->copy (dat); /* copy of dat is the root */
      return;
    }

  tnode = free_loc (btree->tree, btree->cmp, dat);
  if (tnode == NULL) /* dat already in tree or malloc failure */
    return;

  tnode->data = btree->copy (dat);

  return;
}

void*
bintree_get (bintree* btree, void* dat)
{
  node* tnode;
  assert (btree != NULL);

  tnode = find_dat (btree->tree, btree->cmp, dat);
  if (tnode == NULL)
    return; /* not found */

  return btree->copy (tnode->data);
}

void
bintree_remove (bintree* btree, void* dat)
{
  node* tnode;
  assert (btree != NULL);

  if (btree->tree == NULL) 
    return; /* empty tree */

  tnode = del_dat (btree->tree, f, dat);

}



/* defintions of static convience functions */
static node*
free_loc (node* tree_root, cmp_func f, void* dat)
{ 
  if (tree_root == NULL) return NULL;

  switch (f (dat, tree_root->data))
    {
    case 0:
      return NULL;
      break;
    case 1:
      if (tree_root->left == NULL)
	{
	  tree_root->left = new_node ();
	  return tree_root->left;
	}
      /* next call is a tail call and would be eliminated in
	 scheme...but in C the recursion can get very deep and
	 overflow the stack unless the compiler can optimize tail
	 calls (like GCC 3 can with -foptimize-sibling-calls)
       */
      free_loc (tree_root->left, f, dat);
      break;
    case -1:
      if (tree_root->right == NULL)
	{
	  tree_root->right = new_node ();
	  return tree_root->right;
	}
      /* next call is a tail call and would be eliminated in
	 scheme...but in C the recursion can get very deep and
	 overflow the stack unless the compiler can optimize tail
	 calls (like GCC 3 can with -foptimize-sibling-calls)
       */
      free_loc (tree_root->right, f, dat);
      break;
    };
}

static node*
find_dat (node* tree_root, cmp_func f, void* dat)
{ 
  if (tree_root == NULL) return NULL;

  switch (f (dat, tree_root->data))
    {
    case 0:
      return tree_root;
      break;
    case 1:
      if (tree_root->left == NULL) return NULL;
      find_dat (tree_root->left, f, dat);
      break;
    case -1:
      if (tree_root->right == NULL)
	return NULL;
      find_dat (tree_root->right, f, dat);
      break;
    };
}

static node* 
new_node ()
{
  node* temp = malloc (sizeof (node));

  temp->left = NULL;
  temp->right = NULL;
  temp->data = NULL;
}

static node*
del_dat (node* nd, node* parent, cmp_func f, void* dat)
{
  /* returns new root of tree */
  node* root
  if (nd == NULL)
    return NULL;

  if (f (dat, nd->data) == 0) /* node we need to kill */
    {
      node* tmp = nd->left;
      node* smallest = tmp;

      while (tmp && smallest)
	{
	  tmp = smallest;
	  if (tmp->right)
	    smallest = tmp->right;
	}
